{% extends "furo/page.html" %} {% block footer %}
<div class="related-pages">
  {% if next -%}
  <a class="next-page" href="{{ next.link }}">
    <div class="page-info">
      <div class="context">
        <span>{{ _("Next") }}</span>
      </div>
      <div class="title">{{ next.title }}</div>
    </div>
    <svg class="furo-related-icon">
      <use href="#svg-arrow-right"></use>
    </svg>
  </a>
  {%- endif %} {% if prev -%}
  <a class="prev-page" href="{{ prev.link }}">
    <svg class="furo-related-icon">
      <use href="#svg-arrow-right"></use>
    </svg>
    <div class="page-info">
      <div class="context">
        <span>{{ _("Previous") }}</span>
      </div>
      {% if prev.link == pathto(master_doc) %}
      <div class="title">{{ _("Home") }}</div>
      {% else %}
      <div class="title">{{ prev.title }}</div>
      {% endif %}
    </div>
  </a>
  {%- endif %}
</div>
<div class="acknowledgements">
  The Munich Quantum Toolkit has been supported by the European Research Council
  (ERC) under the European Union's Horizon 2020 research and innovation program
  (grant agreement No. 101001318), the Bavarian State Ministry for Science and
  Arts through the Distinguished Professorship Program, as well as the Munich
  Quantum Valley, which is supported by the Bavarian state government with funds
  from the Hightech Agenda Bayern Plus.
  <div style="margin-top: 0.5em">
    <div class="only-light" align="center">
      <img
        src="https://raw.githubusercontent.com/munich-quantum-toolkit/.github/refs/heads/main/docs/_static/mqt-funding-footer-light.svg"
        width="90%"
        alt="MQT Funding Footer"
      />
    </div>
    <div class="only-dark" align="center">
      <img
        src="https://raw.githubusercontent.com/munich-quantum-toolkit/.github/refs/heads/main/docs/_static/mqt-funding-footer-dark.svg"
        width="90%"
        alt="MQT Funding Footer"
      />
    </div>
  </div>
</div>
<div class="bottom-of-page">
  <div class="left-details">
    {%- if show_copyright %}
    <div class="copyright">
      {%- if hasdoc('copyright') %} {% trans path=pathto('copyright'),
      copyright=copyright|e -%}
      <a href="{{ path }}">Copyright</a> &#169; {{ copyright }} {%- endtrans %}
      {%- else %} {% trans copyright=copyright|e -%} Copyright &#169; {{
      copyright }} {%- endtrans %} {%- endif %}
    </div>
    {%- endif %} {%- if last_updated -%}
    <div class="last-updated">
      {% trans last_updated=last_updated|e -%} Last updated on {{ last_updated
      }} {%- endtrans -%}
    </div>
    {%- endif %}
  </div>
  <div class="right-details">
    {% if theme_footer_icons or READTHEDOCS -%}
    <div class="icons">
      {% if theme_footer_icons -%} {% for icon_dict in theme_footer_icons -%}
      <a
        class="muted-link {{ icon_dict.class }}"
        href="{{ icon_dict.url }}"
        aria-label="{{ icon_dict.name }}"
      >
        {{- icon_dict.html -}}
      </a>
      {% endfor %} {%- endif %} {%- endif %}
    </div>
  </div>
  {% endblock footer %}
</div>
